<!DOCTYPE html>
<html lang="en">
    <head>
        <meta charset="utf-8">
        <title>Data-Lunch Error</title>
        <meta name="viewport" content="width=device-width, initial-scale=1">
        <link rel="icon" type="image/x-icon" href="/images/favicon.ico">
        <style>
            /* Import a special font for the title */
            @import url('https://fonts.googleapis.com/css2?family=Silkscreen:wght@700&display=swap');
            @import url('https://fonts.googleapis.com/css2?family=Space+Mono:wght@400&display=swap');
            /* Rules */
            *{
                margin:0;
                padding: 0;
                box-sizing: border-box;
            }
            html{
                height: 100%;
            }
            body{
                font-family: 'Segoe UI', sans-serif;;
                font-size: 1rem;
                line-height: 1.6;
                height: 100%;
            }
            p{
                padding-bottom: 5px;
            }
            .wrap {
                width: 100%;
                height: 100%;
                display: flex;
                justify-content: center;
                align-items: center;
                background: #fafafa;
            }
            .error-frame{
                width: 100%;
                margin: 0 auto;
                border: 1px solid #ddd;
                padding: 2rem;
                background: #ffffff;
                min-width: 250px;
            }
            .error-button{
                background: #0072B5;
                border: 1px solid #ddd;
                color: #ffffff;
                padding: 10px;
                width: 100%;
                text-transform: uppercase;
                max-width: 250px;
                min-width: 85px;
            }
            .error-button:hover{
                background: #005485;
                text-align: center;
            }
            .error-group{
                margin-bottom: 1rem;
                text-align: center;
            }
            .error-header{
                text-align: center;
                margin-bottom : 2rem;
            }
            .error-subtitle{
                font-size: 1rem;
                font-weight: bold;
                text-align: center;
            }
            .error-header h1 {
                font-family: "Silkscreen", cursive;
                font-size: 2rem;
                line-height: 2rem;
            }
            .errormessage {
                font-family: "Space Mono", monospace;
                color: firebrick;
            }
        </style>
    </head>
    <body>
        <div class="wrap">
            <form class="error-frame" action="/logout" method="get">
                <div class="error-header">
                    <h1>Data-Lunch</h1>
                </div>
                <div class="error-subtitle">
                    {{ error_type|default("Error", true) }}: {{ error }}
                </div>
                <div class="error-group">
                    <span class="errormessage">{{error_msg}}</span>
                </div>
                <!--Login Button-->
                <div class="error-group">
                    <button class="error-button" type="submit">Logout</button>
                </div>
            </form>
        </div>
    </body>
</html>